import functools as ft
from collections.abc import Callable
from typing import Any, NoReturn

from beartype.door import is_bearable
from jax import (
    tree as jt,
    tree_util as jtu,
)
from jaxtyping import Array
from jaxtyping._array_types import (
    _anonymous_dim,
    _check_dims,
    _DimType,
    _FixedDim,
    _NamedDim,
    _SymbolicDim,
)
from jaxtyping._storage import get_shape_memo, set_shape_memo

from .error import TypinoxAnnotationError

_vmapped_dim = object()

VMAPPED_LETTER = "$"


def err_array_name(key_path: jtu.KeyPath) -> str:
    """Generates a friendly name from a key path.

    Parameters
    ----------
    key_path : jax.tree_util.KeyPath
        A PyTree key path.

    Returns
    -------
    str
        A human-readable name for the array.
    """
    if len(key_path) == 0:
        return "Array"
    else:
        return f"Array at key path {key_path}"


def single_check_vmapped(
    arr: Array | Any,
    key_path: jtu.KeyPath,
    dim_left: tuple,
    dim_right: tuple,
    single_memo: dict[str, int],
    arg_memo: dict[str, Any],
) -> Array | str | Any:
    """Check a single array for vmapped axes.

    Parameters
    ----------
    arr : Array | Any
        The array to check.
    key_path : jax.tree_util.KeyPath
        The key path of the array, used for error messages.
    dim_left : tuple[jaxtyping._array_types._AbstractDim]
        The dimension specifiers for the first several axes.
    dim_right : tuple[jaxtyping._array_types._AbstractDim]
        The dimension specifiers for the last several axes.
    single_memo : dict[str, int]
        For jaxtyping internal use.
    arg_memo : dict[str, Any]
        For jaxtyping internal use.

    Returns
    -------
    Array | str | Any
        If the array is valid, returns a slice of the array with the vmapped axes removed.
        If the array is invalid, returns an error message.
        If ``arr`` is not an array at all, return it as is.
    """

    if not hasattr(arr, "shape") or not hasattr(arr, "dtype"):
        # Not an array, ignored
        return arr
    shape = arr.shape
    if len(shape) < len(dim_left) + len(dim_right):
        return (
            err_array_name(key_path)
            + " has fewer dimensions than the vmapped axes"
        )
    left_idx = len(dim_left)
    right_idx = len(shape) - len(dim_right)
    if left_idx != 0:
        check = _check_dims(
            dim_left,  # type: ignore
            shape[:left_idx],
            single_memo,
            arg_memo,
        )
        if check != "":
            return err_array_name(key_path) + ": " + check
    if right_idx != len(shape):
        check = _check_dims(
            dim_right,  # type: ignore
            shape[right_idx:],
            single_memo,
            arg_memo,
        )
        if check != "":
            return err_array_name(key_path) + ": " + check
    # Generates a list of slice objects that:
    #   for the first `left_idx` axes, take the 0th element
    #   for the last `right_idx` axes, take the 0th element
    #   for the rest, take all elements
    # This removed the vmapped axes from the array.
    # caveat: doesn't work for axes with size 0
    item_idx = [
        0 if i < left_idx or i >= right_idx else slice(None)
        for i in range(len(shape))
    ]
    return arr[tuple(item_idx)]


def instancecheck_vmapped(
    inner: type,
    dims: tuple[tuple, tuple],
    checker: Callable[[Any, type], bool],
    obj,
    single_memo: dict[str, int],
    arg_memo: dict[str, Any],
) -> str:
    """Check a PyTree structure for vmapped axes.

    Parameters
    ----------
    inner : type
        PyTree structure to check against.
    dims : tuple of two tuples of jaxtyping._array_types._AbstractDim
        The dimension specifiers for the first several axes and the last several axes.
    checker : T -> * -> bool
        An `isinstance`-like function to check the PyTree structure.
    obj : T
        The object to check.
    single_memo : dict[str, int]
        For jaxtyping internal use.
    arg_memo : dict[str, Any]
        For jaxtyping internal use.

    Returns
    -------
    str
        `""` if the object passes the check, or an error message otherwise.
    """

    dim_left, dim_right = dims
    nodedefs, treedef = jt.flatten_with_path(obj)
    leaves: list[Any] = []
    for key_path, arr in nodedefs:
        # the case where `arr` is a string needs to be handled separately...
        if isinstance(arr, str):
            leaves.append(arr)
            continue
        check = single_check_vmapped(
            arr,
            key_path,
            dim_left,
            dim_right,
            single_memo,
            arg_memo,
        )
        # ... as `check` is an str if the single array check fails
        if isinstance(check, str):
            if check != "":
                return check
            check = arr
        leaves.append(check)

    # when all single arrays are valid,
    # map the leaves (without the vmapped axes) back to the PyTree structure
    new_obj = jt.unflatten(treedef, leaves)
    if hasattr(inner, "__instancecheck_str__"):
        return inner.__instancecheck_str__(new_obj)
    if checker(new_obj, inner):
        return ""
    return f"{new_obj} is not an instance of {inner}"


class VmappedMeta(type):
    def __instancecheck__(cls, obj):
        return cls.__instancecheck_str__(obj) == ""

    def __instancecheck_str__(cls, obj) -> str:
        if (
            not hasattr(cls, "inner")
            or not hasattr(cls, "dims")
            or not hasattr(cls, "checker")
        ):
            raise TypinoxAnnotationError(
                "Invalid `Vmapped` class; must have `inner`, `dims` and `checker` attributes."
            )
        single_memo, variadic_memo, pytree_memo, arg_memo = get_shape_memo()
        single_memo_bak = single_memo.copy()
        variadic_memo_bak = variadic_memo.copy()
        pytree_memo_bak = pytree_memo.copy()
        arg_memo_bak = arg_memo.copy()

        try:
            check = instancecheck_vmapped(
                cls.inner,  # type: ignore
                cls.dims,  # type: ignore
                cls.checker,  # type: ignore
                obj,
                single_memo,
                arg_memo,
            )
        except Exception as e:
            set_shape_memo(
                single_memo_bak,
                variadic_memo_bak,
                pytree_memo_bak,
                arg_memo_bak,
            )
            raise e
        if check == "":
            return check
        else:
            set_shape_memo(
                single_memo_bak,
                variadic_memo_bak,
                pytree_memo_bak,
                arg_memo_bak,
            )
            return check

    def __repr__(cls):
        return cls.__module__ + "." + cls.__qualname__

    def __str__(cls):
        return cls.__module__ + "." + cls.__qualname__


class AbstractVmapped(metaclass=VmappedMeta):
    inner: type
    dim_str: str
    dims: tuple[tuple, tuple]
    checker: Callable
    base_name: str

    @classmethod
    def replace_inner(cls, inner):
        return create_vmapped_class(
            cls.base_name, inner, cls.dim_str, cls.dims, cls.checker
        )


@ft.cache
def create_vmapped_class(base_name, inner, dim_str, dims, checker):
    name = f"{base_name}[{inner}, {dim_str}]"
    cls = VmappedMeta(
        name,
        (AbstractVmapped,),
        dict(
            inner=inner,
            dim_str=dim_str,
            dims=dims,
            checker=checker,
            base_name=base_name,
        ),
    )
    cls.__module__ = "typinox"
    return cls


def parse_dims(dim_str: str):
    dims = []
    for index, elem in enumerate(dim_str.split()):
        anonymous = False
        vmapped = False
        if "," in elem and "(" not in elem:
            # Common mistake.
            # Disable in the case that there's brackets to allow for function calls,
            # e.g. `min(foo,bar)`, in symbolic axes.
            raise TypinoxAnnotationError(
                "Axes should be separated with spaces, not commas"
            )
        if elem == "...":
            raise TypinoxAnnotationError(
                "jaxtyping multiple axis not supported in Vmapped; "
                "`...` is not allowed"
            )
        while True:
            if len(elem) == 0:
                # This branch needed as just `_` or `$` is valid
                break
            first_char = elem[0]
            if first_char == "#":
                raise TypinoxAnnotationError(
                    "jaxtyping broadcastable annotation is unsupported in Vmapped; "
                    "`#foo` is not allowed"
                )
            elif first_char == "*":
                raise TypinoxAnnotationError(
                    "jaxtyping multiple axis annotation is unsupported in Vmapped; "
                    "`*foo` is not allowed"
                )
            elif first_char == "_":
                if anonymous:
                    raise TypinoxAnnotationError(
                        "Do not use _ twice to denote anonymity, e.g. `__foo` "
                        "is not allowed"
                    )
                anonymous = True
                elem = elem[1:]
            elif first_char == "?":
                raise TypinoxAnnotationError(
                    "jaxtyping treepath-dependent annotation is unsupported in Vmapped; "
                    "`?foo` is not allowed"
                )
            elif first_char == VMAPPED_LETTER:
                if vmapped:
                    raise TypinoxAnnotationError(
                        "Do not use "
                        + VMAPPED_LETTER
                        + " twice to denote vmapped axes, e.g. `"
                        + VMAPPED_LETTER
                        + VMAPPED_LETTER
                        + "` is not allowed"
                    )
                vmapped = True
                elem = elem[1:]
            elif elem.count("=") == 1:
                _, elem = elem.split("=")
            else:
                break
        elem_size = 0
        if len(elem) == 0 or elem.isidentifier():
            dim_type = _DimType.named
        else:
            try:
                elem_size = int(elem)
            except ValueError:
                dim_type = _DimType.symbolic
            else:
                dim_type = _DimType.fixed

        out: _FixedDim | _NamedDim | _SymbolicDim | object
        if dim_type is _DimType.fixed:
            if anonymous:
                raise TypinoxAnnotationError(
                    "Cannot have a fixed axis be anonymous, e.g. `_4` is not allowed."
                )
            if vmapped:
                raise TypinoxAnnotationError(
                    "Cannot have a fixed axis be vmapped, e.g. `"
                    + VMAPPED_LETTER
                    + "4` is not allowed."
                )
            out = _FixedDim(elem_size, False)
        elif dim_type is _DimType.named:
            if vmapped:
                out = _vmapped_dim
            elif anonymous:
                out = _anonymous_dim
            else:
                out = _NamedDim(elem, False, False)
        else:
            assert dim_type is _DimType.symbolic
            if anonymous:
                raise TypinoxAnnotationError(
                    "Cannot have a symbolic axis be anonymous, "
                    "e.g. `_foo+bar` is not allowed."
                )
            if vmapped:
                raise TypinoxAnnotationError(
                    "Cannot have a symbolic axis be vmapped, "
                    "e.g. `" + VMAPPED_LETTER + "foo+bar` is not allowed."
                )
            out = _SymbolicDim(elem, False)
        dims.append(out)
    n_vmapped_dims = sum(1 for dim in dims if dim is _vmapped_dim)
    if n_vmapped_dims > 1:
        raise TypinoxAnnotationError(
            "Only one axis can be marked as vmapped, e.g. `n "
            + VMAPPED_LETTER
            + " m p` is allowed, but `n "
            + VMAPPED_LETTER
            + " m "
            + VMAPPED_LETTER
            + " p` is not."
        )
    if n_vmapped_dims == 0:
        dims.append(_vmapped_dim)
    if len(dims) == 1:
        return None
    idx = dims.index(_vmapped_dim)
    left, right = dims[:idx], dims[idx + 1 :]
    return tuple(left), tuple(right)


@ft.cache
def make_vmapped(name, inner, dim_str, checker):
    dims = parse_dims(dim_str)
    if dims is None:
        return inner
    dim_left, dim_right = dims
    if (
        inner is not None
        and isinstance(inner, type)
        and issubclass(inner, AbstractVmapped)
    ):
        old_inner, (old_left, old_right) = inner.inner, inner.dims
        dim_left = dim_left + old_left
        dim_right = old_right + dim_right
        inner = old_inner

    return create_vmapped_class(
        name, inner, dim_str, (dim_left, dim_right), checker
    )


def get_vmapped_origin_or_none(cls):
    if not isinstance(cls, type):
        return None
    if issubclass(cls, AbstractVmapped):
        return cls.inner
    return None


class VmappedHelperMeta(type):
    def __instancecheck__(cls, obj) -> NoReturn:
        raise TypinoxAnnotationError(
            "Do not use `Vmapped` as a type hint without specifying"
            " the PyTree structure and the vmapped dimension."
        )

    def __getitem__(cls, params):
        if not isinstance(params, tuple) or len(params) != 2:
            raise TypinoxAnnotationError(
                "Vmapped type hint must be a tuple of a PyTree structure and a string."
            )
        checker = getattr(cls, "checker", isinstance)
        inner, dim_str = params
        dim_str = dim_str.strip()
        return make_vmapped(cls.__qualname__, inner, dim_str, checker)


class VmappedT(metaclass=VmappedHelperMeta):
    """Vmapped type hint.

    When ``VmappedT[inner, shape]`` is used as a type hint,
    the object is expected to be a PyTree structure specified by ``inner``,
    with each array of it extended by ``shape``.
    For example, ``VmappedT[Linear, "device batch"]`` is a type hint for a ``Linear`` PyTree
    with each array shape prepended by two dimensions, named ``device`` and ``batch``.

    **Shape**

    The ``shape`` is a space-separated string of dimension specifiers, such as ``"a 3"``.
    Each dimension can be one of the following:

    - A named dimension, such as ``a``.
    - A fixed dimension, such as ``3``.
    - A symbolic expression in terms of other variable-size axes, such as ``a-1``.
    - ``_``: An anonymous dimension, which is a named dimension that is not matched to any other dimension.

    They function similarly to the dimension specifiers in :mod:`jaxtyping`. It can also be:

    - ``$``: Represents the original array shape before vmapping.
      This is optional, and if omitted, it is assumed that ``$`` appears in the end.

      For example, if ``T`` is ``Float[Array, "3 4"]``,
      then ``Vmapped[T, "batch"]`` and ``Vmapped[T, "batch $"]`` are equivalent to ``Float[Array, "batch 3 4"]``;
      ``Vmapped[T, "$ batch"]`` is equivalent to ``Float[Array, "3 4 batch"]``.

    Other jaxtyping array annotation modifiers, such as ``*``, ``?``, and ``...`` are not supported.

    .. warning::

        None of the axes specified in ``shape`` can be zero (e.g., ``Vmapped[T, "0 3"]`` is invalid).

    **Inner type**

    The ``inner`` is the type hint for a PyTree structure of arrays.
    It can be any valid type hint, including another ``VmappedT`` type hint.

    It is most useful when ``isinstance(x, inner)`` also checks the shape of arrays in ``x``.
    Examples of such ``inner`` hints include:

    - **(jax or numpy) arrays**:
      ``isinstance(x, VmappedT[Array, "b c"])`` checks that
      ``x`` is a jax array with at least two initial dimensions.
    - **tuple of arrays**:
      ``isinstance(x, VmappedT[tuple[Array, Array], "b c"])`` checks that both
      elements of ``x`` are jax arrays with at least two initial dimensions,
      and that their first two dimensions match.
    - **jaxtyping shaped arrays**:
      ``VmappedT[Float[Array, "b c"], "a"]`` is equivalent to ``VmappedT[Array, "a b c"]``,
      while ``VmappedT[Float[Array, "b c"], "$ a"]`` is equivalent to ``VmappedT[Array, "b c a"]``.
    - **Typinox Modules**:
      ``isinstance(x, VmappedT[M, "b c"])`` checks that
      ``x`` is an instance of a :class:`TypedModule <typinox.module.TypedModule>` class ``M``,
      and each array member of ``x`` has at least two initial dimensions.
      If these dimensions are removed, the resulting object should match ``M``.
    """

    def __init__(self) -> NoReturn:
        raise TypinoxAnnotationError(
            "Do not instantiate `VmappedT` directly; use it as a type hint."
        )

    checker = is_bearable


class VmappedI(metaclass=VmappedHelperMeta):
    """Vmapped type hint without using :mod:`beartype`.

    Used only when you do not want to use :func:`is_bearable() <beartype.door.is_bearable>`
    to check the type of the object.
    Useful when the inner annotation is a :mod:`jaxtyping` array type hint.

    The interface is the same as :class:`VmappedT <typinox.vmapped.VmappedT>`.
    """

    def __init__(self) -> NoReturn:
        raise TypinoxAnnotationError(
            "Do not instantiate `VmappedI` directly; use it as a type hint."
        )

    checker = isinstance



import chex
import equinox as eqx
import jax
import numpy as np
import pytest
from beartype import beartype as typechecker
from beartype.door import is_bearable
from jax import (
    numpy as jnp,
)
from jaxtyping import Array, Float, Integer, Scalar, Shaped, jaxtyped
from numpy import ndarray

from typinox.vmapped import VmappedI, VmappedT


def t(a, b):
    return is_bearable(a, b)


@pytest.mark.parametrize("Vmapped", [VmappedI, VmappedT])
def test_vmapped_basic(Vmapped):
    """Test the basic shape handling of Vmapped."""
    arr3 = jnp.array([1, 2, 3])
    arr34 = jnp.arange(12).reshape(3, 4)
    narr34 = np.arange(12).reshape(3, 4)
    arr345 = jnp.arange(60).reshape(3, 4, 5)

    assert t(arr3, Integer[Array, " 3"])
    assert t(arr34, Integer[Array, "3 4"])
    assert t(arr345, Integer[Array, "3 4 5"])

    assert t(arr3, Vmapped[Integer[Scalar, ""], " 3"])
    assert t(arr3, Vmapped[Integer[Array, " 3"], ""])

    assert not t(arr3, Vmapped[Integer[Scalar, ""], " 4"])
    assert not t(arr3, Vmapped[Integer[Array, " 4"], ""])

    assert not t(arr3, Vmapped[Float[Scalar, ""], " 3"])
    assert not t(arr3, Vmapped[Float[Scalar, " 3"], ""])

    assert not t(arr3, Shaped[Array, "3 3"])
    assert not t(arr34, Shaped[Array, "3 3"])
    assert not t(arr3, Shaped[Array, "n n"])
    assert not t(arr34, Shaped[Array, "n n"])

    assert not t(arr3, Vmapped[Shaped[Array, " 3"], " 3"])
    assert not t(arr34, Vmapped[Shaped[Array, " 3"], " 3"])

    assert not t(arr3, Vmapped[Shaped[Array, "n"], " 3"])
    assert not t(arr34, Vmapped[Shaped[Array, "n"], " 4"])
    assert not t(arr34, Vmapped[Shaped[Array, "n"], "m q"])
    assert not t(arr34, Vmapped[Shaped[Array, "n m"], "q"])

    assert t(arr345, Vmapped[Shaped[Array, "n"], "m q"])
    assert t(arr345, Vmapped[Shaped[Array, "n m"], "q"])
    assert not t(arr345, Vmapped[Shaped[Array, "n"], "m"])
    assert not t(arr345, Vmapped[Shaped[Array, ""], "n m q r"])
    assert not t(arr345, Vmapped[Shaped[Array, " n"], "m q r"])
    assert not t(arr345, Vmapped[Shaped[Array, "n m"], "q r"])
    assert not t(arr345, Vmapped[Shaped[Array, "n m q"], " r"])
    assert not t(arr345, Vmapped[Shaped[Array, "n m q r"], ""])

    with jaxtyped("context"):  # type: ignore
        assert not t(arr3, Vmapped[Shaped[Array, " n"], " n"])

    with jaxtyped("context"):  # type: ignore
        assert not t(arr34, Vmapped[Shaped[Array, " n"], " n"])

    with jaxtyped("context"):  # type: ignore
        # n=3, m=4
        assert t(arr3, Integer[Array, " n"])

        assert t(arr3, Vmapped[Integer[Scalar, ""], " n"])
        assert t(arr3, Vmapped[Integer[Array, "n"], ""])

        assert t(arr34, Integer[Array, "n m"])
        assert t(arr34, Vmapped[Integer[Array, "m"], "n"])
        assert t(arr34, Vmapped[Integer[Array, "m"], "n $"])
        assert t(arr34, Vmapped[Integer[Array, "n"], "$ m"])
        assert t(arr34, Vmapped[Integer[Scalar, ""], "n $ m"])

        assert t(arr34, Vmapped[Vmapped[Integer[Scalar, ""], "n $"], "$ m"])

        assert t(narr34, Integer[ndarray, "n m"])
        assert t(narr34, Vmapped[Integer[ndarray, "m"], "n"])
        assert t(narr34, Vmapped[Integer[ndarray, "m"], "n $"])
        assert t(narr34, Vmapped[Integer[ndarray, "n"], "$ m"])


@pytest.mark.parametrize("Vmapped", [VmappedI, VmappedT])
def test_vmapped_large(Vmapped):
    """Test the basic shape handling of Vmapped with more dimensions."""
    arr3345 = jnp.arange(180).reshape(3, 3, 4, 5)
    arr4335 = jnp.einsum("abcd->cbad", arr3345)

    assert t(arr3345, Integer[Array, "3 3 4 5"])
    assert t(arr4335, Integer[Array, "4 3 3 5"])

    assert t(arr3345, Vmapped[Integer[Array, "3 3"], "$ 4 5"])
    assert t(arr4335, Vmapped[Integer[Array, "3 3"], "4 $ 5"])
    assert not t(arr4335, Vmapped[Integer[Array, "3 3"], "$ 4 5"])
    assert not t(arr3345, Vmapped[Integer[Array, "3 3"], "4 $ 5"])

    assert t(arr3345, Vmapped[Integer[Array, "n n"], "$ 4 5"])
    assert t(arr4335, Vmapped[Integer[Array, "n n"], "4 $ 5"])
    assert not t(arr4335, Vmapped[Integer[Array, "n n"], "$ 4 5"])
    assert not t(arr3345, Vmapped[Integer[Array, "n n"], "4 $ 5"])

    assert t(arr3345, Vmapped[Integer[Array, "m q"], "n n $"])
    assert not t(arr4335, Vmapped[Integer[Array, "m q"], "n n $"])

    with jaxtyped("context"):  # type: ignore
        assert t(arr4335, Vmapped[Integer[Array, "n n"], "m $ q"])
    with jaxtyped("context"):  # type: ignore
        assert not t(arr3345, Vmapped[Integer[Array, "n n"], "m $ q"])

    with jaxtyped("context"):  # type: ignore
        assert t(arr4335, Vmapped[Integer[Array, "n q"], "m n $"])
    with jaxtyped("context"):  # type: ignore
        assert not t(arr3345, Vmapped[Integer[Array, "n q"], "m n $"])

    with jaxtyped("context"):  # type: ignore
        assert t(arr4335, Vmapped[Integer[Array, "n q"], "m n"])
    with jaxtyped("context"):  # type: ignore
        assert not t(arr3345, Vmapped[Integer[Array, "n q"], "m n"])

    with jaxtyped("context"):  # type: ignore
        assert t(arr4335, Vmapped[Integer[Array, "m n"], "$ n q"])
    with jaxtyped("context"):  # type: ignore
        assert not t(arr3345, Vmapped[Integer[Array, "m n"], "$ n q"])


@pytest.mark.parametrize("Vmapped", [VmappedT])
def test_vmapped_vmap_simple(Vmapped):
    """Test Vmapped with vmap using a basic function."""

    @eqx.filter_jit
    def inner(x: Float[Scalar, ""]) -> tuple[str, Float[Array, "3"]]:
        y = x * jnp.ones(3)
        assert t(y, Float[Array, "3"])
        return "aaa", y

    xs = jnp.sin(jnp.arange(5))
    ys = eqx.filter_vmap(inner)(xs)
    with jaxtyped("context"):  # type: ignore
        assert t(xs, Vmapped[inner.__annotations__["x"], "n"])
        assert t(ys, Vmapped[inner.__annotations__["return"], "n"])

    xss = jnp.sin(jnp.arange(6).reshape(2, 3))
    yss = eqx.filter_vmap(eqx.filter_vmap(inner))(xss)
    with jaxtyped("context"):  # type: ignore
        assert t(xss, Vmapped[inner.__annotations__["x"], "m n"])
        assert t(yss, Vmapped[inner.__annotations__["return"], "m n"])
    with jaxtyped("context"):  # type: ignore
        assert not t(xss, Vmapped[inner.__annotations__["x"], "n n"])
        assert not t(yss, Vmapped[inner.__annotations__["return"], "n n"])

    xss = jnp.sin(jnp.arange(9).reshape(3, 3))
    yss = eqx.filter_vmap(eqx.filter_vmap(inner))(xss)
    with jaxtyped("context"):  # type: ignore
        assert t(xss, Vmapped[inner.__annotations__["x"], "n n"])
        assert t(yss, Vmapped[inner.__annotations__["return"], "n n"])


@pytest.mark.parametrize("Vmapped", [VmappedI, VmappedT])
def test_vmapped_vmap_axes(Vmapped):
    """Test Vmapped with vmap using a function with in_axes and out_axes."""

    @jax.jit
    def inner(
        x: Float[Scalar, ""], y: Float[Array, " n"]
    ) -> Float[Array, "n n"]:
        z = (y[:, None] + x) * y[None, :]
        return jnp.sin(z) / x

    x = jnp.array(2.0)
    ys = jnp.arange(12).astype(float).reshape(3, 4)

    zs = jax.vmap(inner, in_axes=(None, 0))(x, ys)
    with jaxtyped("context"):  # type: ignore
        assert t(x, Vmapped[inner.__annotations__["x"], ""])
        assert t(ys, Vmapped[inner.__annotations__["y"], "m"])
        assert t(zs, Vmapped[inner.__annotations__["return"], "m"])

    zs = jax.vmap(inner, in_axes=(None, 1), out_axes=2)(x, ys)
    with jaxtyped("context"):  # type: ignore
        assert t(x, Vmapped[inner.__annotations__["x"], ""])
        assert t(ys, Vmapped[inner.__annotations__["y"], "$ m"])
        assert t(zs, Vmapped[inner.__annotations__["return"], "$ m"])

    zs = jax.vmap(inner, in_axes=(None, -1), out_axes=-1)(x, ys)
    with jaxtyped("context"):  # type: ignore
        assert t(x, Vmapped[inner.__annotations__["x"], ""])
        assert t(ys, Vmapped[inner.__annotations__["y"], "$ m"])
        assert t(zs, Vmapped[inner.__annotations__["return"], "$ m"])

    xs = jnp.arange(5).astype(float)
    ys = jnp.arange(4).astype(float)
    zs = jax.vmap(inner, in_axes=(0, None), out_axes=0)(xs, ys)
    with jaxtyped("context"):  # type: ignore
        assert t(xs, Vmapped[inner.__annotations__["x"], "m"])
        assert t(ys, Vmapped[inner.__annotations__["y"], ""])
        assert t(zs, Vmapped[inner.__annotations__["return"], "m"])


ArrayOfTwo = Float[Array, " 2"]


def test_doc_example():
    @jaxtyped(typechecker=typechecker)
    def my_function(x: ArrayOfTwo, y: ArrayOfTwo) -> ArrayOfTwo:
        return x + y

    @jaxtyped(typechecker=typechecker)
    def my_vmapped(
        x: VmappedT[ArrayOfTwo, " n"], y: VmappedT[ArrayOfTwo, " n"]
    ) -> VmappedT[ArrayOfTwo, " n"]:
        return jax.vmap(my_function)(x, y)

    chex.assert_trees_all_close(
        my_vmapped(jnp.ones((3, 2)), jnp.ones((3, 2))), jnp.ones((3, 2)) * 2
    )

    with pytest.raises(TypeError):
        my_vmapped(jnp.ones((3, 2)), jnp.ones((4, 2)))



from typing import Any

import chex
import jax
from beartype.door import is_bearable
from jax import numpy as jnp
from jaxtyping import Array, Shaped

import typinox as tpx
from typinox import Vmapped


def my_is_bearable(x: Any, T) -> bool:  # noqa: F821
    """To make pyright happy."""
    return is_bearable(x, T)


def test_unstack_simple():
    t = jnp.arange(6).reshape(3, 2)
    u = jnp.arange(45).reshape(3, 5, 3)
    l = (t, u)
    s = list(tpx.tree.unstack(l))
    assert len(s) == 3
    assert my_is_bearable(
        s, list[tuple[Shaped[Array, " 2"], Shaped[Array, " 5 3"]]]
    )


def test_stack_simple():
    l = [
        (
            jnp.arange(4).reshape(2, 2),
            {"a": jnp.arange(3).reshape(1, 1, 3, 1, 1)},
        )
        for _ in range(5)
    ]
    s = tpx.tree.stack(l)
    assert s[0].shape == (5, 2, 2)
    assert s[1]["a"].shape == (5, 1, 1, 3, 1, 1)
    assert my_is_bearable(
        s,
        Vmapped[
            tuple[
                Shaped[Array, " 2 2"], dict[str, Shaped[Array, " 1 1 3 1 1"]]
            ],
            "n",
        ],
    )


def test_stack_unstack():
    def f(x):
        return {
            "a": x,
            "b": [
                jnp.arange(3).reshape(1, 1, 3, 1, 1) + x,
                (jnp.arange(12).reshape(2, 3, 1, 2) - x).astype(jnp.float16),
                (
                    (jnp.arange(6) - x).reshape(2, 3),
                    jnp.sin(jnp.arange(10)) * x,
                ),
            ],
        }

    xs = jnp.arange(10)
    xl = [float(x) for x in range(10)]

    ys = jax.vmap(f)(xs)
    yl = [f(x) for x in xl]

    chex.assert_trees_all_equal(ys, tpx.tree.stack(tpx.tree.unstack(ys)))
    chex.assert_trees_all_equal(yl, list(tpx.tree.unstack(tpx.tree.stack(yl))))

    chex.assert_trees_all_equal(ys, tpx.tree.stack(yl))
    chex.assert_trees_all_equal(yl, list(tpx.tree.unstack(ys)))



from typing import Callable  # noqa: UP035

__all__: list[str] = []


def func_to_bracket(fn: Callable | None = None, name: str | None = None):
    if fn is None:
        return lambda fn: func_to_bracket(fn, name)
    if name is None:
        name = fn.__name__
    cls = type(name, (), {"__class_getitem__": lambda cls, item: fn(item)})
    cls.__module__ = fn.__module__
    return cls



from beartype.typing import TYPE_CHECKING, Annotated, TypeAliasType, Unpack
from beartype.vale import Is

from ._helper import func_to_bracket

UnpackType = type(Unpack[tuple[int, str]])


class ValidationFailed(ValueError):
    pass


class ValidateProto:
    def __validate__(self) -> bool:
        return True

    def __validate_str__(self) -> str:
        return ""


def validate_str_single(cls, obj) -> str:
    if cls.__dict__.get("__validate_str__", None) is not None:
        validated = cls.__validate_str__(obj)
        if validated != "":
            return validated
    if cls.__dict__.get("__validate__", None) is not None:
        try:
            validated = cls.__validate__(obj)
        except ValidationFailed as e:
            return str(e)
        if validated is False:
            return "the custom validation failed"
    return ""


def validate_str(obj) -> str:
    cls = type(obj)
    for kls in cls.__mro__[
        :-1
    ]:  # skip the last class in the mro, which is object
        validated = validate_str_single(kls, obj)
        if validated != "":
            return validated
    return ""


def _validate(obj):
    return validate_str(obj) == ""


TypinoxValid = Is[_validate]
if TYPE_CHECKING:
    type ValidatedT[_T] = Annotated[_T, TypinoxValid]
else:

    @func_to_bracket
    def ValidatedT[T](cls: type[T]) -> type[T]:
        from ._vmapped import AbstractVmapped

        if isinstance(cls, UnpackType):
            # workaround for weird Python typing restriction
            # see: https://github.com/beartype/beartype/issues/562#issuecomment-3404611632
            wrap_cls = TypeAliasType(cls.__name__, cls)
            return Annotated[wrap_cls, TypinoxValid]
        if not isinstance(cls, type):
            return Annotated[cls, TypinoxValid]
        if issubclass(cls, AbstractVmapped):
            return cls.replace_inner(ValidatedT[cls.inner])
        if (
            getattr(cls, "__validate__", None) is None
            and getattr(cls, "__validate_str__", None) is None
        ):
            return cls
        return Annotated[cls, TypinoxValid]



import dataclasses
from typing import Any, Protocol, overload


@dataclasses.dataclass(frozen=True)
class _FakeArray:
    shape: tuple[int, ...]
    dtype: Any = dataclasses.field(default=None)


class NDArray(Protocol):
    """
    Duck match for any object that behaves like a numpy array
    (has ``.shape`` and ``.dtype`` attributes).
    """

    @property
    def shape(self) -> tuple[int, ...]: ...

    @property
    def dtype(self) -> Any: ...


ShapeLike = int | tuple[int, ...] | NDArray


def shape_sanitize(shape: ShapeLike) -> tuple[int, ...]:
    if isinstance(shape, int):
        return (shape,)
    if isinstance(shape, tuple):
        return shape
    return shape.shape


@overload
def ensure_shape(shape: ShapeLike, dim_spec: str, /): ...


@overload
def ensure_shape(name: str, shape: ShapeLike, dim_spec: str, /): ...


def ensure_shape(*args):
    """
    Ensure that the shape of an array matches :mod:`jaxtyping` named dimensions.

    Parameters
    ----------
    name: str, optional
        The name of the array.

    shape : ShapeLike
        The shape of the array.

    dim_spec : str
        The :mod:`jaxtyping` named dimensions to be matched.

    Returns
    -------
    None
        If the shape matches the named dimensions, return None.

    Raises
    ------
    ValidationFailed
        when the shape does not match the named dimensions
    """

    from jaxtyping import Shaped

    from .validator import ValidationFailed

    if len(args) == 2:
        shape, dim_spec = args
        name = ""
    elif len(args) == 3:
        name, shape, dim_spec = args
    else:
        raise ValueError(f"invalid number of arguments: {len(args)}")

    obj = _FakeArray(shape_sanitize(shape))
    if not isinstance(obj, Shaped[_FakeArray, dim_spec]):  # type: ignore
        if name:
            raise ValidationFailed(
                f'{name} has shape {shape} which does not match the named dimensions "{dim_spec}"'
            )
        else:
            raise ValidationFailed(
                f"shape {shape} does not match the named dimensions {dim_spec}"
            )


@overload
def ensure_shape_equal(shape1: ShapeLike, shape2: ShapeLike, /): ...


@overload
def ensure_shape_equal(name: str, shape1: ShapeLike, shape2: ShapeLike, /): ...


@overload
def ensure_shape_equal(
    name1: str, shape1: ShapeLike, name2: str, shape2: ShapeLike, /
): ...


def ensure_shape_equal(*args):
    """
    Ensure that the shapes of two arrays are equal.

    Parameters
    ----------
    name : str, optional
        The name of the arrays. Cannot be provided if name1 or name2 is provided.

    name1 : str, optional
        The name of the first array. Must be provided if name2 is provided.

    shape1 : int, or tuple[int, ...], or an object with ``.shape``
        The shape of the first array.

    name2 : str, optional
        The name of the second array. Must be provided if name1 is provided.

    shape2 : int, or tuple[int, ...], or an object with ``.shape``
        The shape of the second array.

    Returns
    -------
    None
        If the shapes are equal, return None.

    Raises
    ------
    ValidationFailed
        when the shapes are not equal.
    """

    from .validator import ValidationFailed

    if len(args) == 2:
        shape1, shape2 = args
        name1 = name2 = ""
    elif len(args) == 3:
        name, shape1, shape2 = args
        name1 = name2 = name
    elif len(args) == 4:
        name1, shape1, name2, shape2 = args
    else:
        raise ValueError(f"invalid number of arguments: {len(args)}")

    if shape_sanitize(shape1) != shape_sanitize(shape2):
        raise ValidationFailed(
            f"{name1} {shape1} does not match {name2} {shape2}"
        )



from typing import Self
from unittest.mock import Mock

from beartype.door import die_if_unbearable
from beartype.typing import Annotated, Any, Callable
from beartype.vale import Is

from typinox import (
    TypedModule,
    ValidatedT,
    set_debug_mode,
)

set_debug_mode(True)


def test_diamond_inheritance():
    quacks = [Mock(return_value=True) for _ in range(4)]
    happys = [Mock(return_value=True) for _ in range(4)]

    def happy_func(i: int) -> Callable[[Any], bool]:
        def z(x):
            happys[i](x)
            return True

        return z

    result_mock = Mock(return_value=True)

    def good_result(x):
        return result_mock(x)

    class A(TypedModule):
        a: Annotated[int, Is[happy_func(0)]]

        def __validate__(self):
            return self.a >= 0 and quacks[0]()

    class B(A):
        b: Annotated[int, Is[happy_func(1)]]

        def __validate__(self):
            return self.b >= 0 and quacks[1]()

    class C(A):
        c: Annotated[int, Is[happy_func(2)]]

        def __validate__(self):
            return self.c >= 0 and quacks[2]()

    class D(B, C):
        d: Annotated[int, Is[happy_func(3)]]

        def __validate__(self):
            return self.d >= 0 and quacks[3]()

        def value(self: Self) -> Annotated[int, Is[good_result]]:
            return self.a + self.b + self.c + self.d

    obj = D(a=9, b=12, c=233, d=7)

    for q in quacks:
        q.assert_called_once()
        q.reset_mock()
    for h, num in zip(happys, [9, 12, 233, 7]):
        h.assert_called_once_with(num)
        h.reset_mock()

    die_if_unbearable(obj, ValidatedT[D])

    for q in quacks:
        q.assert_called_once()
        q.reset_mock()
    for h, num in zip(happys, [9, 12, 233, 7]):
        h.assert_called_once_with(num)
        h.reset_mock()

    assert obj.value() == 9 + 12 + 233 + 7

    # Because jaxtyping likes to call every parameter twice
    for q in quacks:
        assert q.call_count == 2
        q.reset_mock()
    for h, num in zip(happys, [9, 12, 233, 7]):
        assert h.call_count == 2
        h.assert_called_with(num)
        h.reset_mock()
    result_mock.assert_called_once()
    result_mock.reset_mock()



import pytest

from typinox import (
    AbstractVar,
    TypedModule,
    set_debug_mode,
)

set_debug_mode(True)


class Base(TypedModule):
    a: int
    b: AbstractVar[int]


def test_abstract_var_basic():
    with pytest.raises(TypeError):
        _ = Base(a=1, b=1)


class SubWithValue(Base):
    b: int  # type: ignore


def test_abstract_var_subclass():
    obj = SubWithValue(a=1, b=3)
    assert obj.a == 1
    assert obj.b == 3


class SubWithProperty(Base):
    @property  # type: ignore
    def b(self) -> int:  # type: ignore
        return self.a + 2


def test_abstract_var_property():
    obj = SubWithProperty(a=1)  # type: ignore
    assert obj.a == 1
    assert obj.b == 3



from typing import Any, Self

import jax
import pytest
from beartype.door import is_bearable
from jax import (
    numpy as jnp,
    tree as jt,
)
from jaxtyping import Float, Scalar, jaxtyped

import typinox as tpx
from typinox import (
    TypedModule,
    ValidatedT,
    Vmapped,
    set_debug_mode,
)
from typinox.tree import unstack

set_debug_mode(True)


def my_is_bearable(x: Any, T) -> bool:
    """To make pyright happy."""
    return is_bearable(x, T)


class SinWave(TypedModule):
    freq: Float[Scalar, ""]
    phase: Float[Scalar, ""]
    amp: Float[Scalar, ""]

    def __call__(self, x: Float[Scalar, ""]) -> Float[Scalar, ""]:
        return self.amp * jnp.sin(self.freq * x + self.phase)

    def rotate_90_degs_bad(self) -> Vmapped[Self, "4"]:
        return self

    def actual_rotate_90_degs(self):
        def rotate(x):
            return self.__class__(
                freq=self.freq,
                phase=self.phase + jnp.pi / 4 * x,
                amp=self.amp,
            )

        return jax.vmap(rotate)(jnp.arange(4))

    def rotate_90_degs_bad_sig(self) -> Self:
        return self.actual_rotate_90_degs()

    def rotate_90_degs_good(self) -> Vmapped[Self, "4"]:
        return self.actual_rotate_90_degs()


SinWaveT = ValidatedT[SinWave]


def test_simple_vmap():
    freqs = jnp.array([1.0, 2.0, 3.0])
    phases = jnp.array([0.0, 1.0, 2.0]) / jnp.pi
    amps = jnp.array([0.5, 1.0, 1.5])
    waves = jax.vmap(SinWave)(freqs, phases, amps)
    assert isinstance(waves, SinWave)
    assert not my_is_bearable(waves, SinWaveT)
    assert my_is_bearable(waves, Vmapped[SinWaveT, "3"])

    nums = jnp.array([1.0, 2.0, 3.0])

    def call(wave, num):
        return wave(num)

    _ = jax.vmap(call)(waves, nums)
    _ = jax.vmap(call, in_axes=(0, None))(waves, nums[0])


def test_vmapped_self():
    sw = SinWave(jnp.array(2.0), jnp.array(4.0), jnp.array(6.0))
    with pytest.raises(TypeError):
        sw.rotate_90_degs_bad()

    with pytest.raises(TypeError):
        sw.rotate_90_degs_bad_sig()

    sws = sw.rotate_90_degs_good()

    swss = jax.vmap(lambda x: x.rotate_90_degs_good())(sws)
    assert swss.freq.shape == (4, 4)


class MultiWaves(TypedModule):
    n: int = tpx.field(static=True)
    waves: tuple[SinWaveT, ...]
    shifts: tuple[Float[Scalar, ""], ...]

    def __validate__(self) -> bool:
        return len(self.waves) == self.n and len(self.shifts) == self.n

    def __call__(self, x: Float[Scalar, ""]) -> Float[Scalar, ""]:
        for wave, shift in zip(self.waves, self.shifts):
            x = wave(x) + shift
        return x


MultiWavesT = ValidatedT[MultiWaves]


def test_multi_waves():
    def create_multiwave(k):
        n = 4
        freqs = jnp.array([1.0, 2.0, 3.0, 4.0]) + k
        phases = jnp.array([0.0, 1.0, 2.0, 3.0]) / jnp.pi
        amps = jnp.array([0.5, 1.0, 1.5, 2.0 - k])
        waves = jax.vmap(SinWave)(freqs, phases, amps)
        shifts = jnp.array([0.0, 1.0, -1.0, 0.0]) * k
        return MultiWaves(n, tuple(unstack(waves)), tuple(shifts))

    mw = create_multiwave(jnp.array(3.0))
    assert my_is_bearable(mw, MultiWavesT)
    assert my_is_bearable(mw(jnp.array(4.0)), Float[Scalar, ""])

    leaves, treedef = jt.flatten(mw)
    leaves[-1] = jnp.array([4.0, 3.0])
    bad_mw = jt.unflatten(treedef, leaves)
    assert not all([my_is_bearable(bad_mw, MultiWavesT) for _ in range(30)])

    with pytest.raises(TypeError):
        mw(jnp.array([4.0, 3.0]))

    with pytest.raises(TypeError):
        _ = MultiWaves(3, mw.waves, mw.shifts)

    mws = jax.vmap(create_multiwave)(jnp.array([1.0, 2.0, 3.0]))
    assert isinstance(mws, MultiWaves)
    assert not my_is_bearable(mws, MultiWavesT)
    assert my_is_bearable(mws, Vmapped[MultiWavesT, "c"])


class SegmentTree(TypedModule):
    n: int = tpx.field(static=True)
    left: Self | None
    right: Self | None
    value: Float[Scalar, ""]
    total: Float[Scalar, ""]

    def __validate__(self) -> bool:
        if self.n == 1:
            return self.left is None and self.right is None
        left_size = self.left.n if self.left is not None else 0
        right_size = self.right.n if self.right is not None else 0
        return (
            self.n == left_size + right_size + 1
            and -1 <= left_size - right_size <= 1
        )

    @classmethod
    def create_leaf(cls, value) -> Self:
        return cls(1, None, None, value, value)

    @classmethod
    def create_node(cls, left, value, right):
        return cls(
            (left.n if left else 0) + 1 + (right.n if right else 0),
            left,
            right,
            value,
            (left.total if left else 0.0)
            + value
            + (right.total if right else 0.0),
        )

    @classmethod
    def build(cls, values) -> Self | None:
        n = len(values)
        if n == 0:
            return None
        if n == 1:
            return cls.create_leaf(values[0])
        mid = n // 2
        left = cls.build(values[:mid])
        right = cls.build(values[mid + 1 :])
        return cls.create_node(left, values[mid], right)

    def query(self, i, j):
        if i < 0:
            i = 0
        if j >= self.n:
            j = self.n
        if i >= j:
            return jnp.array(0.0)
        if i <= 0 <= self.n <= j:
            return self.total
        left_size = self.left.n if self.left is not None else 0
        left_value = (
            self.left.query(i, j) if self.left is not None else jnp.array(0.0)
        )
        right_value = (
            self.right.query(i - left_size - 1, j - left_size - 1)
            if self.right is not None
            else jnp.array(0.0)
        )
        return (
            left_value
            + right_value
            + (self.value if i <= left_size < j else 0.0)
        )


SegmentTreeT = ValidatedT[SegmentTree]


def test_segment_tree():
    arr = jnp.sin(jnp.arange(20))
    st = SegmentTree.build(arr)
    assert st is not None
    assert my_is_bearable(st, SegmentTreeT)

    for i, j in [(3, 7), (6, 16), (0, 15), (-3, 37)]:
        st_value = st.query(i, j)
        cons_arr = (jnp.arange(20) >= i) & (jnp.arange(20) < j)
        arr_value = jnp.sum(arr * cons_arr)
        assert jnp.allclose(st_value, arr_value)

    def create_st(k, z):
        arr = jnp.sin(jnp.arange(20) * k + z)
        return SegmentTree.build(arr)

    def elim_st(st):
        return tuple(
            [st.query(i, j) for i, j in [(3, 7), (6, 16), (0, 15), (-3, 37)]]
        )

    sts = jax.vmap(create_st)(
        jnp.array([1.0, 2.0, 3.0]), jnp.array([0.0, 1.0, 2.0])
    )
    assert isinstance(sts, SegmentTree)
    assert not my_is_bearable(sts, SegmentTreeT)
    sums = jax.vmap(elim_st)(sts)

    with jaxtyped("context"):  # type: ignore
        assert my_is_bearable(sts, Vmapped[SegmentTreeT, "c"])
        assert my_is_bearable(
            sums, Vmapped[tuple[(Float[Scalar, ""],) * 4], "c"]
        )

    with jaxtyped("context"):  # type: ignore
        assert my_is_bearable(sts, Vmapped[SegmentTreeT, "c"])
        assert not my_is_bearable(
            sums, Vmapped[tuple[(Float[Scalar, ""],) * 4], "c+1"]
        )



import chex
from beartype.typing import Self
from jax import (
    numpy as jnp,
)

from typinox import (
    TypedModule,
    set_debug_mode,
)

set_debug_mode(True)


class MyTuple[*Ts](TypedModule):
    values: tuple[*Ts]

    def set(self, *args: *Ts) -> Self:
        return self.__class__(values=args)


def test_my_tuple():
    t = MyTuple(values=(jnp.array([1.0, 2.0]), jnp.array([[1.0], [2.0]])))
    assert isinstance(t.values, tuple)
    assert len(t.values) == 2
    chex.assert_trees_all_close(t.values[0], jnp.array([1.0, 2.0]))
    chex.assert_trees_all_close(t.values[1], jnp.array([[1.0], [2.0]]))

    t1 = t.set(jnp.array([3.0, 4.0]), jnp.array([[3.0], [4.0]]))
    assert isinstance(t1.values, tuple)
    assert len(t1.values) == 2
    chex.assert_trees_all_close(t1.values[0], jnp.array([3.0, 4.0]))
    chex.assert_trees_all_close(t1.values[1], jnp.array([[3.0], [4.0]]))

    # The original instance is unchanged.
    assert isinstance(t.values, tuple)
    assert len(t.values) == 2
    chex.assert_trees_all_close(t.values[0], jnp.array([1.0, 2.0]))
    chex.assert_trees_all_close(t.values[1], jnp.array([[1.0], [2.0]]))



from dataclasses import dataclass
from typing import Any

import beartype
import beartype.roar
import pytest
from beartype import beartype as typechecker
from beartype.door import is_bearable

from typinox import ValidatedT, ValidationFailed


def my_is_bearable(x: Any, T) -> bool:
    """To make pyright happy."""
    return is_bearable(x, T)


class ClassA:
    def __init__(self, x: int, y: int):
        if x > 0:
            self.x = x
        self.y = y

    def __validate__(self):
        if hasattr(self, "x"):
            return True
        else:
            if self.y > 0:
                return False
            else:
                raise ValidationFailed("the x attribute is missing")


class ClassBBase:
    def __init__(self, x: int, y: int):
        if x > 0:
            self.x = x
        self.y = y

    def __validate_str__(self):
        if hasattr(self, "x"):
            return ""
        else:
            return "this instance has no x attribute"


class ClassB(ClassBBase):
    pass


@dataclass
class ClassC:
    x: int
    y: int


def fn_a_vanilla(a: ClassA) -> int:
    return a.x + a.y


def fn_a_custom(a: ValidatedT[ClassA]) -> int:
    return a.x + a.y


fn_a_typechecked = typechecker(fn_a_vanilla)
fn_a_custom_typechecked = typechecker(fn_a_custom)

fn_a = (fn_a_vanilla, fn_a_custom, fn_a_typechecked, fn_a_custom_typechecked)


def fn_b_vanilla(b: ClassB) -> int:
    return b.x + b.y


def fn_b_custom(b: ValidatedT[ClassB]) -> int:
    return b.x + b.y


fn_b_typechecked = typechecker(fn_b_vanilla)
fn_b_custom_typechecked = typechecker(fn_b_custom)

fn_b = (fn_b_vanilla, fn_b_custom, fn_b_typechecked, fn_b_custom_typechecked)


@pytest.mark.parametrize("cls,fn", [(ClassA, fn_a), (ClassB, fn_b)])
def test_good_case(cls, fn):
    fn_vanilla, fn_custom, fn_typechecked, fn_custom_typechecked = fn
    assert fn_vanilla(cls(1, 2)) == 3
    assert fn_custom(cls(1, 2)) == 3
    assert fn_typechecked(cls(1, 2)) == 3
    assert fn_custom_typechecked(cls(1, 2)) == 3


@pytest.mark.parametrize(
    "cls,fn,errorstr",
    [
        (ClassA, fn_a, "custom validation failed"),
        (ClassB, fn_b, "this instance has no x attribute"),
    ],
)
def test_bad_attribute_read(cls, fn, errorstr):
    fn_vanilla, fn_custom, fn_typechecked, fn_custom_typechecked = fn
    with pytest.raises(AttributeError):
        fn_vanilla(cls(0, 2))

    with pytest.raises(AttributeError):
        fn_custom(cls(0, 2))

    with pytest.raises(AttributeError):
        fn_typechecked(cls(0, 2))

    with pytest.raises(
        beartype.roar.BeartypeCallHintParamViolation,
        # match=errorstr,
    ):
        fn_custom_typechecked(cls(0, 2))


@pytest.mark.parametrize(
    "cls,fn,errorstr",
    [
        (ClassA, fn_a, "the x attribute is missing"),
        (ClassB, fn_b, "this instance has no x attribute"),
    ],
)
def test_bad_attribute_read_pretty(cls, fn, errorstr):
    fn_vanilla, fn_custom, fn_typechecked, fn_custom_typechecked = fn
    with pytest.raises(AttributeError):
        fn_vanilla(cls(0, 0))

    with pytest.raises(AttributeError):
        fn_custom(cls(0, 0))

    with pytest.raises(AttributeError):
        fn_typechecked(cls(0, 0))

    with pytest.raises(
        beartype.roar.BeartypeCallHintParamViolation,
        # match=errorstr,
    ):
        fn_custom_typechecked(cls(0, 0))


@pytest.mark.parametrize("cls,fn", [(ClassA, fn_a), (ClassB, fn_b)])
def test_bad_type(cls, fn):
    fn_vanilla, fn_custom, fn_typechecked, fn_custom_typechecked = fn
    assert fn_vanilla(ClassC(1, 2)) == 3  # type: ignore
    assert fn_custom(ClassC(1, 2)) == 3  # type: ignore

    with pytest.raises(beartype.roar.BeartypeCallHintParamViolation):
        fn_typechecked(ClassC(1, 2))  # type: ignore

    with pytest.raises(beartype.roar.BeartypeCallHintParamViolation):
        fn_custom_typechecked(ClassC(1, 2))  # type: ignore


def test_validator_unrelated():
    assert my_is_bearable(1, ValidatedT[int])
    assert not my_is_bearable(1, ValidatedT[str])


def test_validator_union():
    assert not my_is_bearable(1, ValidatedT[ClassA | ClassB])
    assert my_is_bearable(ClassA(1, 2), ValidatedT[ClassA | ClassB])
    assert my_is_bearable(ClassB(2, 1), ValidatedT[ClassA | ClassB])
    assert not my_is_bearable(ClassA(-1, 2), ValidatedT[ClassA | ClassB])
    assert not my_is_bearable(ClassB(-2, 1), ValidatedT[ClassA | ClassB])



import chex
from jax import (
    numpy as jnp,
    random as jr,
)
from jaxtyping import Array, Float, Key, Scalar

from typinox import (
    TypedModule,
    set_debug_mode,
)

set_debug_mode(True)


class GaussianGenerator(TypedModule):
    mean: Float[Array, " d"]
    linear: Float[Array, "d d"]

    def __init__(self, mean: Float[Array, " d"], cov: Float[Array, "d d"]):
        self.mean = mean
        self.linear = jnp.linalg.cholesky(cov, upper=True)

    def sample(
        self, key: Key[Scalar, ""], shape: tuple[int, ...] = ()
    ) -> Float[Array, "... d"]:
        z = jr.normal(key, shape + self.mean.shape)
        return self.mean.reshape(
            (1,) * len(shape) + self.mean.shape
        ) + jnp.einsum("...i,ij->...j", z, self.linear)


def test_gaussian_generator():
    key = jr.key(0)
    mean = jnp.array([1.0, 2.0])
    cov = jnp.array([[1.0, 0.5], [0.5, 2.0]])
    gen = GaussianGenerator(mean, cov)

    assert isinstance(gen.mean, jnp.ndarray)
    assert isinstance(gen.linear, jnp.ndarray)
    assert gen.mean.shape == (2,)
    assert gen.linear.shape == (2, 2)

    samples = gen.sample(key, shape=(10000,))
    assert samples.shape == (10000, 2)

    # Check that the samples have approximately the correct mean and covariance
    sample_mean = jnp.mean(samples, axis=0)
    sample_cov = jnp.cov(samples, rowvar=False)

    chex.assert_trees_all_close(sample_mean, mean, atol=0.1)
    chex.assert_trees_all_close(sample_cov, cov, atol=0.1)



import jaxtyping
from beartype.typing import Generator, Iterable
from jax import (
    numpy as jnp,
    tree as jt,
)

from .vmapped import Vmapped

PyTreeDef = type[jaxtyping.PyTreeDef]  # to make mypy happy


def stack[T](trees: Iterable[T]) -> Vmapped[T, " n"]:
    """Stacks every corresponding leaf on a list of PyTrees.

    For example, given two trees ((a, b), c) and ((a', b'), c'), returns
    ((stack(a, a'), stack(b, b')), stack(c, c')).
    Useful for turning a list of objects into something you can feed to a
    vmapped function.

    Parameters
    ----------
    T : type, type parameter
        The type of the PyTrees
    trees : Iterable[T]
        A list of PyTrees, all with the same structure ``T``

    Returns
    -------
    Vmapped[T, "n"]
        A PyTree with the same structure as the input trees, but with the leaves stacked


    Examples
    --------

    >>> a = ({ "l": jnp.array([1, 2]), "r": jnp.array(3) },
    ...      jnp.array([4, 5, 6]))
    >>> b = ({ "l": jnp.array([7, 8]), "r": jnp.array(9) },
    ...      jnp.array([10, 11, 12]))
    >>> c = typinox.tree.stack([a, b])
    >>> d = ({ "l": jnp.array([[1, 2], [7, 8]]), "r": jnp.array([3, 9]) },
    ...      jnp.array([[4, 5, 6], [10, 11, 12]]))
    >>> chex.assert_trees_all_equal(c, d)

    """
    leaves_list = []
    treedef_list: list[PyTreeDef] = []
    for tree in trees:
        leaves, treedef = jt.flatten(tree)
        leaves_list.append(leaves)
        treedef_list.append(treedef)

    grouped_leaves = zip(*leaves_list)
    result_leaves = [jnp.stack(l) for l in grouped_leaves]
    return treedef_list[0].unflatten(result_leaves)


def unstack[T](tree: Vmapped[T, " _"]) -> Generator[T]:
    """Unstacks a PyTree of arrays into a list of PyTrees. Inverse of :func:`stack`.

    For example, given a tree ((a, b), c), where a, b, and c all have first
    dimension k, will make k trees
    [((a[0], b[0]), c[0]), ..., ((a[k], b[k]), c[k])].

    Useful for turning the output of a vmapped function into normal objects.

    Parameters
    ----------
    T : type, type parameter
        The type of the PyTrees
    tree : Vmapped[T, " _"]
        A PyTree of structure ``T`` with the leaves stacked

    Yields
    ------
    Generator[T]
        A generator that yields PyTrees with the same structure ``T`` as the input tree,
        but with the leaves unstacked

    Examples
    --------

    >>> a = ({ "l": jnp.array([[1, 2], [3, 4]]), "r": jnp.array([5, 6]) },
    ...      jnp.array([7, 8]))
    >>> aa = list(typinox.tree.unstack(a))
    >>> bb = [
    ...       ({ "l": jnp.array([1, 2]), "r": jnp.array(5) },
    ...        jnp.array(7)),
    ...       ({ "l": jnp.array([3, 4]), "r": jnp.array(6) },
    ...        jnp.array(8))
    ...      ]
    >>> chex.assert_trees_all_equal(aa, bb)
    """
    treedef: PyTreeDef  # to make mypy happy
    leaves, treedef = jt.flatten(tree)
    n_trees = leaves[0].shape[0]
    for i in range(n_trees):
        new_leaves = [leaf[i] for leaf in leaves]
        yield treedef.unflatten(new_leaves)



import warnings

debug_mode = False


class TypinoxUnknownTypeWarning(UserWarning):
    pass


class TypinoxUnknownFunctionWarning(UserWarning):
    pass


def set_debug_mode(value: bool = True):
    global debug_mode
    debug_mode = value


def debug_print(*args, **kwargs):
    if debug_mode:
        print(*args, **kwargs)


def debug_warn(*args, **kwargs):
    if debug_mode:
        warnings.warn(*args, **kwargs)


def debug_raise(err: type[Exception], *args, **kwargs):
    if debug_mode:
        raise err(*args, **kwargs)



import dataclasses
import inspect
import warnings
import weakref
from abc import ABCMeta
from types import FunctionType

import beartype
from beartype.door import is_bearable
from beartype.typing import (
    Annotated,
    Any,
    Callable,
    Iterable,
    Never,
    Self,
    Sequence,
    Unpack,
    cast,
    get_args,
    get_origin,
    overload,
)
from equinox import (
    AbstractClassVar,
    AbstractVar,
    field as eqx_field,
)
from equinox._module._module import (
    _has_dataclass_init,
    _ModuleMeta as EqxModuleMeta,
)
from equinox._module._prebuilt import BoundMethod as EqxWrapMethod
from jaxtyping import jaxtyped

from ._vmapped import (
    AbstractVmapped,
    VmappedMeta,
    get_vmapped_origin_or_none,
)
from .debug import (
    TypinoxUnknownFunctionWarning,
    debug_warn,
)
from .error import TypinoxNotImplementedError, TypinoxTypeViolation
from .shaped import ensure_shape as ensure_shape
from .validator import ValidatedT, ValidationFailed, validate_str

AnnotatedAlias = cast(type, type(Annotated[int, ">3"]))
CallableAliasType = type(Callable[[int], float])
GenericAliasType = type(tuple[int, str])
UnpackType = type(Unpack[tuple[int, str]])
UnionType = type(int | float)
UnionGenericAlias = type(Self | None)


@overload
def field(): ...


@overload
def field(
    *,
    typecheck: bool = True,
    converter: Callable[[Any], Any] = ...,
    static: bool = False,
    default: Any = ...,
    default_factory: Callable[[], Any] | Any = ...,
    init: bool = True,
    hash: bool | None = None,
    metadata: dict[str, Any] | None = None,
    kw_only: bool = ...,
): ...


def field(
    *,
    typecheck: bool = True,
    metadata: dict[str, Any] | None = None,
    **kwargs,
) -> dataclasses.Field:
    """Specify a field of a typinox typed module.

    Parameters
    ----------

    typecheck : bool, default True
        If specified as False, the field will be ignored during typechecking.

    converter : Callable[[Any], Any], optional
        Used by ``__init__`` to pre-process the value.
        See :func:`equinox.field`.

    static : bool, default False
        If specified as True, the field will be static. This will make it
        a part of the pytree structure, instead of a subtree or a leaf.
        See :func:`equinox.field`.

    default : Any, optional
        Default value used by ``__init__`` if this field is not provided.
        See :func:`dataclasses.field`.

    default_factory : Callable[[], Any] | Any, optional
        Used to generate a default value by ``__init__`` if this field is not provided.
        See :func:`dataclasses.field`.

    metadata : dict[str, Any] | None, optional
        Metadata to be attached to the field.
    """
    if metadata is None:
        metadata = {}
    metadata["typecheck"] = typecheck
    return eqx_field(
        metadata=metadata,
        **kwargs,
    )


@dataclasses.dataclass(frozen=True)
class TypedPolicy:
    """Used to configure the typechecking behavior of a module.

    As an example, if you want to disable typechecking for a specific method,
    you can use the following:

    .. code-block:: python

        class MyModule(TypedModule,
                    typed_policy=TypedPolicy(skip_methods={"wtf"})):
            a: int
            b: int = field(typecheck=False)

            def wtf(self, x: int):
                return x

        z = MyModule(1, "not an int")
        z.wtf("also not an int")

    .. hint::

        You can also use the :func:`typing.no_type_check` decorator
        to disable typechecking for a specific method.

    Parameters
    ----------

    always_validated : bool, default True
        If True, every argument of methods will be type checked with a
        custom validator if present.

    typecheck_init_result : bool, default True
        If False, type checking will not be performed on the result
        of ``__init__``.

    skip_methods : frozenset[str], default {}
        Specifies which methods should skip type checking.
    """

    always_validated: bool = dataclasses.field(default=True)
    typecheck_init_result: bool = dataclasses.field(default=True)
    skip_methods: frozenset[str] = dataclasses.field(default_factory=frozenset)

    def __init__(
        self,
        always_validated: bool = True,
        typecheck_init_result: bool = True,
        skip_methods: Iterable[str] = frozenset(),
    ):
        object.__setattr__(self, "always_validated", always_validated)
        object.__setattr__(self, "typecheck_init_result", typecheck_init_result)
        object.__setattr__(self, "skip_methods", frozenset(skip_methods))


policy_for_type: weakref.WeakKeyDictionary[type, TypedPolicy] = (
    weakref.WeakKeyDictionary()
)


def mark_as_typed[T: Callable](fn: T) -> T:
    if getattr(fn, "__typinox_typed__", False):
        return fn
    setattr(fn, "__typinox_typed__", True)
    return fn


def marked_as_typed(fn: Callable) -> bool:
    return getattr(fn, "__typinox_typed__", False)


def decorate_function(fn: Callable) -> Callable:
    return jaxtyped(fn, typechecker=beartype.beartype)


def fold_or(args: Sequence[Any]) -> Any:
    if len(args) == 0:
        return Never
    result = args[0]
    for arg in args[1:]:
        result = result | arg
    return result


def sanitize_annotation(annotation: Any, cls: type) -> Any:
    """Recursively sanitize the annotation.

    Replaces ``Self`` with the class itself; recurses into ``Union`` and
    similar types."""
    if annotation is Self:
        return cls
    if isinstance(annotation, UnionType | UnionGenericAlias):
        args = get_args(annotation)
        return fold_or([sanitize_annotation(arg, cls) for arg in args])
    if isinstance(annotation, UnpackType):
        args = get_args(annotation)
        assert len(args) == 1
        inner = args[0]
        return Unpack[sanitize_annotation(inner, cls)]
    if isinstance(annotation, GenericAliasType):
        if isinstance(annotation, CallableAliasType):
            return annotation
        origin = get_origin(annotation)
        args = get_args(annotation)
        if origin is None:
            raise TypinoxNotImplementedError(
                f"Unsupported GenericAlias: {annotation}"
            )
        return origin[tuple(sanitize_annotation(arg, cls) for arg in args)]
    if isinstance(annotation, VmappedMeta):
        origin = get_vmapped_origin_or_none(annotation)
        if origin is Self:
            return cast(type[AbstractVmapped], annotation).replace_inner(cls)
    if isinstance(annotation, AnnotatedAlias):
        origin = getattr(annotation, "__origin__")
        if origin is Self:
            return cls
        return Annotated[
            sanitize_annotation(origin, cls),
            *getattr(annotation, "__metadata__", []),
        ]
    return annotation


def sanitize_member_annotation(annotation: Any, cls: type) -> Any:
    if get_origin(annotation) is AbstractVar:
        annotation_args = get_args(annotation)
        assert len(annotation_args) == 1
        inner_annotation = sanitize_annotation(annotation_args[0], cls)
        return inner_annotation | property
    if get_origin(annotation) is AbstractClassVar:
        raise TypinoxNotImplementedError(
            "AbstractClassVar is not yet supported by Typinox."
        )
    return sanitize_annotation(annotation, cls)


def method_transform_annotations(
    fn: FunctionType, cls: type, policy: TypedPolicy
) -> FunctionType:
    """Sanitize all annotations of a method.

    Also wraps every annotation with ValidatedT[]."""
    annotations = fn.__annotations__
    for key, value in annotations.items():
        if isinstance(value, str):
            warnings.warn(
                f"Typinox: string annotations are not supported: `{value}` in {fn} of {cls}"
            )
            continue
        new_annotation = sanitize_annotation(value, cls)
        if policy.always_validated:
            new_annotation = ValidatedT[new_annotation]  # type: ignore
        if new_annotation is not value:
            annotations[key] = new_annotation
    return fn


def is_magic(name: str) -> bool:
    return name.startswith("__") and name.endswith("__")


CallableDescriptor = staticmethod | classmethod | property | EqxWrapMethod

SKIP_MAGIC_MODULES = frozenset(
    [
        "builtins",
        "typing",
        "dataclasses",
        "typinox",
        "typinox.module",
        "typinox._module",
        "equinox",
        "equinox._module",
    ]
)
SKIP_MAGIC_NAMES = frozenset(
    [
        "__validate__",
        "__validate_str__",
        "__validate_self_str__",
    ]
)


def skip_magic(
    name: str, fn: Callable | CallableDescriptor, cls: type, policy: TypedPolicy
) -> bool:
    if name in SKIP_MAGIC_NAMES:
        return True
    if fn.__module__ in SKIP_MAGIC_MODULES:
        return True
    return False


def decorate_method[T: Callable | CallableDescriptor](
    name: str, fn: T, cls: type, policy: TypedPolicy
) -> T:
    """Decorate a method with the typechecker (jaxtyped and beartype).
    Recurses into staticmethods, classmethods and properties."""
    if name in policy.skip_methods:
        return fn
    if getattr(fn, "__no_type_check__", False):
        return fn
    if isinstance(fn, staticmethod):
        actual_method = fn.__func__
        return cast(
            T, staticmethod(decorate_method(name, actual_method, cls, policy))
        )
    if isinstance(fn, classmethod):
        actual_method = fn.__func__
        return cast(
            T, classmethod(decorate_method(name, actual_method, cls, policy))
        )
    if isinstance(fn, property):
        fget = (
            decorate_method(name, fn.fget, cls, policy)
            if fn.fget is not None
            else None
        )
        fset = (
            decorate_method(name, fn.fset, cls, policy)
            if fn.fset is not None
            else None
        )
        fdel = (
            decorate_method(name, fn.fdel, cls, policy)
            if fn.fdel is not None
            else None
        )
        return cast(T, property(fget, fset, fdel))
    # # not needed after equinox 0.13
    # if isinstance(fn, EqxWrapMethod):
    #     return cast(
    #         T, EqxWrapMethod(decorate_method(name, fn.method, cls, policy))
    #     )
    if not callable(fn):
        return fn
    if not inspect.isfunction(fn):
        # We can only wrap Python-native functions.
        debug_warn(
            f"Typinox: attempting to perform typechecking decoration on unknown object: {fn}",
            TypinoxUnknownFunctionWarning,
        )
        return cast(T, fn)
    if marked_as_typed(fn):
        return cast(T, fn)
    if is_magic(name):
        if skip_magic(name, fn, cls, policy):
            return cast(T, fn)
    # Main case: pure-python function.
    pyfunc = cast(FunctionType, fn)
    pyfunc = method_transform_annotations(pyfunc, cls, policy)
    decorated = decorate_function(pyfunc)
    decorated = mark_as_typed(decorated)
    return cast(T, decorated)


class RealTypedModuleMeta(EqxModuleMeta):
    """Metaclass for TypedModule.

    If you want to create a module with a metaclass other than
    :class:`abc.ABCMeta`, :class:`equinox._module._ModuleMeta` or
    :class:`typinox.module.TypedModuleMeta`, you need to create a
    new metaclass that inherits from this class and your metaclass of
    choice.
    """

    def __new__(
        mcs,
        name,
        bases,
        dict_,
        /,
        strict: bool | None = False,
        typed_policy: TypedPolicy | dict | None = None,
        **kwargs,
    ):
        # [Step 1] Create the Module as normal.
        cls = super().__new__(mcs, name, bases, dict_, strict=strict, **kwargs)
        # Assumption:
        # - Every non-magic normal method is wrapped by Equinox.
        # - A __init__ method is created, either by the user or by Equinox.

        # [Step 2] Wrap all methods with the typechecker.
        # [Step 2.0] Prepare the typechecking policy.
        if isinstance(typed_policy, dict):
            typed_policy = TypedPolicy(**typed_policy)
        if typed_policy is None:
            typed_policy = TypedPolicy()
        policy_for_type[cls] = typed_policy
        # [Step 2.1] Wrap the methods with the typechecker.
        for key, value in cls.__dict__.items():
            if key == "__init__":
                # We skip __init__ method generated by Equinox.
                if _has_dataclass_init[cls]:
                    continue
            decorated_value = decorate_method(key, value, cls, typed_policy)
            if decorated_value is not value:
                setattr(cls, key, decorated_value)

        # [Step 3] Add the validator methods.
        old_validate = cls.__dict__.get("__validate__", None)
        old_validate_str = cls.__dict__.get("__validate_str__", None)

        # [Step 3.1] Recursively validate the fields.
        # [Step 3.1.0] Prepare the annotations to check.
        sanitized_annotations = {
            key: sanitize_member_annotation(value, cls)
            for key, value in cls.__annotations__.items()
        }
        # Exclude the fields that are marked as not typechecking.
        for field in dataclasses.fields(cls):
            if not field.metadata.get("typecheck", True):
                sanitized_annotations.pop(field.name, None)

        # [Step 3.1 cont'd] Actually validate the fields.
        def __validate_self_str__(self):
            __tracebackhide__ = True
            for member, hint in sanitized_annotations.items():
                if member not in self.__dict__:
                    continue
                value = self.__dict__[member]
                if not is_bearable(value, hint):
                    return f"its {member} does not match type hint {hint}, got {value}"
            if old_validate_str is not None:
                result = old_validate_str(self)
                if result:
                    return result
            if old_validate is not None:
                try:
                    result = old_validate(self)
                except ValidationFailed as e:
                    return str(e)
                if result is False:
                    return "it failed its custom validation"
            return ""

        def __validate_str__(self):
            __tracebackhide__ = True
            with jaxtyped("context"):  # type: ignore
                return __validate_self_str__(self)

        # Add the methods to the class.
        __validate_str__.__qualname__ = "__validate_str__"
        ABCMeta.__setattr__(cls, "__validate_self_str__", __validate_self_str__)
        ABCMeta.__setattr__(cls, "__validate_str__", __validate_str__)
        ABCMeta.__setattr__(cls, "__validate__", None)
        return cls

    # Creating an instance with MyModule(...) will call this method.
    def __call__(cls, *args, **kwargs):
        __tracebackhide__ = True
        # [Step 1] Create the instance as normal.
        instance = super().__call__(*args, **kwargs)
        # [Step 2] Typecheck the instance.
        policy = policy_for_type[cls]
        if policy.typecheck_init_result:
            check_result = validate_str(instance)
            if check_result:
                raise TypinoxTypeViolation(
                    f"The instance {instance} of {cls} has failed typechecking, as {check_result}"
                )
        return instance


RealTypedModuleMeta.__name__ = "TypedModuleMeta"



from typing import Self

import chex
import numpy as np
import pytest
from beartype.door import die_if_unbearable
from jax import (
    numpy as jnp,
    tree as jt,
)
from jaxtyping import Array, Float, Scalar

import typinox as tpx
from typinox import (
    TypedModule,
    ValidatedT,
    set_debug_mode,
)

set_debug_mode(True)


class Point2D(TypedModule):
    x: Float[Scalar, ""]
    y: Float[Scalar, ""]

    def norm_correct_1(self) -> Float[Scalar, ""]:
        return (self.x**2 + self.y**2) ** 0.5

    norm_correct_1_1 = norm_correct_1

    def norm_correct_2(self) -> Float[Array, "..."]:
        return (self.x**2 + self.y**2) ** 0.5

    def norm_correct_3(self):
        return (self.x**2 + self.y**2) ** 0.5

    def norm_incorrect_1(self) -> float:
        return (self.x**2 + self.y**2) ** 0.5  # type: ignore

    def norm_incorrect_2(self) -> Float[Array, " n"]:
        return (self.x**2 + self.y**2) ** 0.5

    def swap_correct(self) -> Self:
        return self.__class__(x=self.y, y=self.x)

    def swap_incorrect(self) -> Float[Array, " 2"]:
        return self.__class__(x=self.y, y=self.x)  # type: ignore


Point2D_T = ValidatedT[Point2D]


def test_point2d():
    with pytest.raises(TypeError):
        Point2D_T(x=3.0, y=4.0)  # type: ignore

    p = Point2D_T(x=jnp.float32(3.0), y=jnp.float32(4.0))
    assert 4.99 < p.norm_correct_1() < 5.01
    assert 4.99 < p.norm_correct_1_1() < 5.01
    assert 4.99 < p.norm_correct_2() < 5.01
    assert 4.99 < p.norm_correct_3() < 5.01

    with pytest.raises(TypeError):
        p.norm_incorrect_1()

    with pytest.raises(TypeError):
        p.norm_incorrect_2()

    assert p.swap_correct().x == 4.0
    assert p.swap_correct().y == 3.0

    with pytest.raises(TypeError):
        p.swap_incorrect()


class TerriblePoint2D(TypedModule):
    # Why, oh why
    x: Float[Scalar, ""] = tpx.field(typecheck=False)
    y: Float[Scalar, ""] = tpx.field(typecheck=False)

    def norm_whatever(self):
        return (self.x**2 + self.y**2) ** 0.5


TerriblePoint2D_T = ValidatedT[TerriblePoint2D]


def test_field_untypecheck():
    p = TerriblePoint2D(x=jnp.float32(3.0), y=jnp.float32(4.0))
    assert 4.99 < p.norm_whatever() < 5.01
    die_if_unbearable(p, TerriblePoint2D_T)

    p = TerriblePoint2D(x=3, y=4)  # type: ignore
    assert 4.99 < p.norm_whatever() < 5.01
    die_if_unbearable(p, TerriblePoint2D_T)

    p = TerriblePoint2D(x=np.float16(3), y=np.int64(4))  # type: ignore
    assert 4.99 < p.norm_whatever() < 5.01
    die_if_unbearable(p, TerriblePoint2D_T)


class Point3D(Point2D):
    z: Float[Scalar, ""]

    def norm_correct_1(self) -> Float[Scalar, ""]:
        return (self.x**2 + self.y**2 + self.z**2) ** 0.5

    def norm_incorrect_1(self) -> float:
        return (self.x**2 + self.y**2) ** 0.5  # type: ignore

    def cycle_correct_1(self) -> Self:
        return self.__class__(x=self.y, y=self.z, z=self.x)

    def cycle_correct_2(self) -> Self:
        leaves, treedef = jt.flatten(self)
        return jt.unflatten(treedef, leaves[1:] + leaves[:1])

    def cycle_incorrect(self) -> Self:
        leaves, treedef = jt.flatten(self)
        return jt.unflatten(treedef, [float(leaf) for leaf in leaves])


def test_subclassing():
    p = Point3D(
        x=jnp.float32(3.0 * 4), y=jnp.float32(4.0 * 4), z=jnp.float32(5.0 * 3)
    )
    assert 24.99 < p.norm_correct_1() < 25.01

    with pytest.raises(TypeError):
        p.norm_incorrect_1()

    assert p.cycle_correct_1().x == 4.0 * 4
    assert p.cycle_correct_2().x == 4.0 * 4


class SquareMat(TypedModule):
    data: Float[Array, "n n"]

    def transpose_correct(self) -> Self:
        return self.__class__(data=self.data.T)

    def transpose_danger_but_correct(self):
        return jt.map(lambda x: x[0], self)

    def transpose_incorrect_1(self) -> Self:
        return self.transpose_danger_but_correct()

    def transpose_incorrect_selfparam(self: Self):
        return jt.map(lambda x: x[0], self.data)


def test_self_check():
    p = Point3D(
        x=jnp.float32(3.0 * 4), y=jnp.float32(4.0 * 4), z=jnp.float32(5.0 * 3)
    )
    with pytest.raises(TypeError):
        p.cycle_incorrect()

    m = SquareMat(data=jnp.eye(3))
    assert m.transpose_correct().data.shape == (3, 3)
    assert m.transpose_danger_but_correct().data.shape == (3,)

    with pytest.raises(TypeError):
        m.transpose_incorrect_1()

    bad_m = m.transpose_danger_but_correct()
    with pytest.raises(TypeError):
        bad_m.transpose_incorrect_selfparam()


def test_doc_example():
    class AffineMap(TypedModule):  # also known as linear layer
        k: Float[Array, "n m"]
        b: Float[Array, " n"]

        def __call__(self, x: Float[Array, " m"]) -> Float[Array, " n"]:
            self._validate()  # populates jaxtyping shape storage
            return jnp.dot(self.k, x) + self.b

        def compose(self, other: Self) -> Self:  # Self annotation is supported!
            return self.__class__(
                k=jnp.dot(self.k, other.k), b=self.b + jnp.dot(self.k, other.b)
            )

    f1 = AffineMap(k=jnp.arange(6).reshape((3, 2)).astype(float), b=jnp.ones(3))
    f2 = AffineMap(k=jnp.ones((5, 3)) / 18, b=jnp.ones(5))

    chex.assert_trees_all_close(f1(jnp.ones(2)), jnp.array([2.0, 6.0, 10.0]))

    with pytest.raises(TypeError):
        f1(jnp.ones(3))

    f3 = f2.compose(f1)
    chex.assert_trees_all_close(f3(jnp.ones(2)), jnp.ones(5) * 2)



from . import tree as tree
from .debug import (
    debug_print as debug_print,
    debug_raise as debug_raise,
    debug_warn as debug_warn,
    set_debug_mode as set_debug_mode,
)
from .module import (
    AbstractVar as AbstractVar,
    TypedModule as TypedModule,
    TypedPolicy as TypedPolicy,
    field as field,
)
from .shaped import (
    ensure_shape as ensure_shape,
    ensure_shape_equal as ensure_shape_equal,
)
from .validator import (
    ValidatedT as ValidatedT,
    ValidationFailed as ValidationFailed,
)
from .vmapped import (
    Vmapped as Vmapped,
    VmappedI as VmappedI,
    VmappedT as VmappedT,
)

Module = TypedModule



from typing import Self

import jax
import pytest

from typinox import (
    TypedModule,
    TypedPolicy,
)


def test_no_typecheck():
    from typing import no_type_check

    class Module(TypedModule):
        @no_type_check
        def f(self, x: int) -> int:
            return x

    obj = Module()
    obj.f("not an int")  # type: ignore


def test_skip_method():
    policy = TypedPolicy(skip_methods=["f"])

    class Module(TypedModule, typed_policy=policy):
        def f(self, x: int) -> int:
            return x

    obj = Module()
    obj.f("not an int")  # type: ignore


def test_not_always_validated():
    policy = TypedPolicy(always_validated=False)

    class Module(TypedModule, typed_policy=policy):
        x: int

        def f(self, rhs: Self) -> None:
            pass

    obj = Module(2)
    rhs = jax.tree.map(lambda x: str(x), obj)
    obj.f(rhs)

    class ControlGroup(TypedModule):
        x: int

        def f(self, rhs: Self) -> None:
            pass

    obj = ControlGroup(2)
    rhs = jax.tree.map(lambda x: str(x), obj)
    with pytest.raises(TypeError):
        obj.f(rhs)


def test_no_typecheck_init_result():
    policy = TypedPolicy(typecheck_init_result=False)

    class Module(TypedModule, typed_policy=policy):
        x: int

        def __init__(self, x: int) -> None:
            super().__init__()
            self.x = str(x)  # type: ignore

    obj = Module(2)
    assert obj.x == "2"

    class ControlGroup(TypedModule):
        x: int

        def __init__(self, x: int) -> None:
            super().__init__()
            self.x = str(x)  # type: ignore

    with pytest.raises(TypeError):
        obj = ControlGroup(2)



import jax
from beartype.typing import Callable
from jax import (
    numpy as jnp,
    random as jr,
    tree as jt,
)
from jaxtyping import Array, Float, Scalar

import typinox as tpx
from typinox import (
    TypedModule,
    ValidatedT,
    Vmapped,
    ensure_shape,
    set_debug_mode,
)

set_debug_mode(True)


class LinearLayer(TypedModule):
    w: Float[Array, "in out"]
    b: Float[Array, " out"]

    def __call__(self, x: Float[Array, " in"]) -> Float[Array, " out"]:
        return jnp.dot(x, self.w) + self.b


LinearLayerT = ValidatedT[LinearLayer]


class ActivationLayer(TypedModule):
    fn: Callable[[Float[Array, " in"]], Float[Array, " out=in"]] = tpx.field(
        static=True
    )

    def __call__(self, x: Float[Array, " in"]) -> Float[Array, " out=in"]:
        return self.fn(x)


ActivationLayerT = ValidatedT[ActivationLayer]


class Sequential(TypedModule):
    layers: list[LinearLayerT | ActivationLayerT]

    def __call__(self, x: Float[Array, " in"]) -> Float[Array, " out"]:
        for layer in self.layers:
            x = layer(x)
        return x


batch_size = 1000


def loss(net, dk) -> Scalar:
    xy = jr.uniform(dk, (batch_size, 2), minval=-1, maxval=1)
    res = jax.vmap(net)(xy).reshape(-1)
    truth = (xy[:, 1] * xy[:, 0] < 0).astype(jnp.float32)
    return jnp.mean((res - truth) ** 2)


@jax.jit
def train_step(net, dk, lr):
    def iter(net, sk):
        l, grad = jax.value_and_grad(loss)(net, sk)
        new_net = jt.map(lambda n, g: n - lr * g, net, grad)
        return new_net, l

    new_net, l = jax.lax.scan(iter, net, jr.split(dk, 500), unroll=2)
    return new_net, jnp.mean(l)


def test_end2end_train():
    key = jr.key(0)
    key11, key12, key21, key22, key31, key32, key = jr.split(key, 7)
    net = Sequential(
        [
            LinearLayer(jr.normal(key11, (2, 3)), jr.normal(key12, (3,))),
            ActivationLayer(jnp.tanh),
            LinearLayer(jr.normal(key21, (3, 4)), jr.normal(key22, (4,))),
            ActivationLayer(jnp.tanh),
            LinearLayer(jr.normal(key31, (4, 1)), jr.normal(key32, (1,))),
            ActivationLayer(jax.nn.sigmoid),
        ]
    )
    data_key, key = jr.split(key, 2)

    for dk in jr.split(data_key, 1):
        net, l = train_step(net, dk, 1.0)

    l = loss(net, key)

    assert l < 0.25 * 0.2


class MixtureOfExperts(TypedModule):
    n_experts: int = tpx.field(static=True)
    experts: Vmapped[Sequential, " n_experts"]
    gate: Sequential

    def __validate__(self):
        ensure_shape(".n_experts", self.n_experts, "n_experts")

    def __call__(self, x: Float[Array, " in"]) -> Float[Array, " out"]:
        g = self.gate(x)
        choice = jax.vmap(lambda expert, weight: weight * expert(x))(
            self.experts, g
        )
        return jnp.sum(choice, axis=0)


MixtureOfExpertsT = ValidatedT[MixtureOfExperts]


def test_end2end_train_moe():
    def gen_expert(k):
        k1, k2, k3, k4 = jr.split(k, 4)
        return Sequential(
            [
                LinearLayer(jr.normal(k1, (2, 3)), jr.normal(k2, (3,))),
                ActivationLayer(jnp.tanh),
                LinearLayer(jr.normal(k3, (3, 1)), jr.normal(k4, (1,))),
            ]
        )

    def gen_gate(k, m):
        k1, k2, k3, k4 = jr.split(k, 4)
        return Sequential(
            [
                LinearLayer(jr.normal(k1, (2, 2)), jr.normal(k2, (2,))),
                ActivationLayer(jax.nn.sigmoid),
                LinearLayer(jr.normal(k3, (2, m)), jr.normal(k4, (m,))),
                ActivationLayer(jax.nn.softmax),
            ]
        )

    def gen_moe(k, m) -> MixtureOfExpertsT:
        k1, k2 = jr.split(k, 2)
        return MixtureOfExperts(
            m, jax.vmap(gen_expert)(jr.split(k1, m)), gen_gate(k2, m)
        )

    key = jr.key(1)
    gen_key, data_key, key = jr.split(key, 3)
    moe = gen_moe(gen_key, 3)

    for dk in jr.split(data_key, 3):
        moe, l = train_step(moe, dk, 1.0)

    l = loss(moe, key)
    assert l < 0.25 * 0.2



import dataclasses
from typing import TYPE_CHECKING, dataclass_transform

import equinox
import equinox._module

from ._module import (
    TypedPolicy as TypedPolicy,
    field as field,
)
from .error import TypinoxTypeViolation

if TYPE_CHECKING:
    # Workaround for static type checkers

    type AbstractVar[T] = T | property

    @dataclass_transform(
        field_specifiers=(dataclasses.field, equinox.field, field),
    )
    class TypedModuleMeta(equinox._module._ModuleMeta):  # type: ignore
        pass

    class TypedModule(equinox.Module, metaclass=TypedModuleMeta):
        def __validate_self_str__(self) -> str:
            return ""

        def _validate(self) -> None: ...

else:
    # The real branch; this is what gets executed at runtime

    from ._module import RealTypedModuleMeta

    AbstractVar = equinox.AbstractVar
    TypedModuleMeta = RealTypedModuleMeta

    class TypedModule(equinox.Module, metaclass=TypedModuleMeta):
        """
        Base class. Inherit this to create a typinox typed module.

        This class is a subclass of :class:`equinox.Module` and provides
        automatic type validation for its fields.

        Thanks to :class:`equinox.Module`, every typed module is
        automatically a `dataclass`_ and a `pytree`_.

        .. _dataclass: https://docs.python.org/3/library/dataclasses.html
        .. _pytree: https://docs.jax.dev/en/latest/pytrees.html

        **Fields**

        Declare typed fields at the class level, using the same syntax as
        for `dataclass`_. The class will automatically become a pytree
        containing all the fields. It will also automatically validate
        the types of the fields at runtime.

        .. code-block:: python

            from jaxtyping import Array, Float, Key

            class MyModule(TypedModule):
                weight: Float[Array, "m n"]
                bias: Float[Array, "m"]
                sublayers: list[TypedModule]

        **Initialization**

        A default constructor is provided similar to a dataclass. It fills
        each field with the arguments in order. You can also pass keyword
        arguments to fill the fields of the same name. For example
        ``MyModule(w, sublayers=[m1, m2], bias=b)``.

        Alternatively, you can provide an ``__init__`` method to customize
        the initialization behavior.

        .. code-block:: python

            from jaxtyping import Array, Float, Key

            class MyModule(TypedModule):
                weight: Float[Array, "m n"]
                bias: Float[Array, "m"]
                sublayers: list[TypedModule]

                def __init__(self, m: int, n: int,
                             key: Key[Array, ""],
                             sublayers: list[TypedModule] = [],
                            ):
                    self.weight = jax.random.normal(key, (m, n))
                    self.bias = jax.ones((m,))
                    self.sublayers = sublayers

        **Methods**

        Define methods at the class level just like any other class.
        Every methods is automatically wrapped with :func:`beartype.beartype`
        and :func:`jaxtyping.jaxtyped` to perform run-time type checking.

        .. code-block:: python

            class MyModule(TypedModule):
                # ... same as above

                def __call__(self, x: Float[Array, "n"]) -> Float[Array, "m"]:
                    y = jnp.dot(self.weight, x) + self.bias
                    for layer in self.sublayers:
                        y = layer(y)
                    # if y is not a Float[Array, "m"] at this point,
                    # an error will be raised by beartype
                    return y

        .. tip::

            The method does not have to be named ``__call__()``; it might
            as well be named ``forward()``. The dunder name ``__call__``
            is not specially treated by typinox.
        """

        pass

    def _validate(self) -> None:
        """A helper method to validate the type of the module.

        This is particularly useful for modules that may be vmapped.
        The argument to and return value of :func:`jax.vmap` are
        pytrees of arrays with added dimensions, and thus may be invalid
        modules. For example:

        .. code-block:: python

            def create_module(key: Key[Array, ""]):
                return MyModule(n=3, m=2, key=key, sublayers=[])

            key = jax.random.key(1)
            some_keys = jax.random.split(key, 5)
            some_modules = jax.vmap(create_module)(some_keys)

        In this case, ``some_modules`` has ``MyModule`` as its ``.__class__``,
        but is not a valid ``MyModule`` because its ``weight`` and ``bias``
        have shapes ``(5, 2, 3)`` and ``(5, 2)`` respectively. Therefore,
        ``some_modules._validate()`` will fail.

        .. hint::

            In this case we can annotate it with :class:`Vmapped`, as it
            passes the type check for ``Vmapped[MyModule, "5"]``.

        Returns
        -------
        None

        Raises
        ------
        TypinoxTypeViolation
            If the type of the module is not valid.
        """
        __tracebackhide__ = True
        cls = type(self)
        for kls in cls.__mro__[-2::-1]:
            if hasattr(kls, "__validate_self_str__"):
                validated = kls.__validate_self_str__(self)
                if validated != "":
                    raise TypinoxTypeViolation(
                        f"the value ({self}) is not a {cls}, as {validated}"
                    )

    # Patch the _validate method to the class
    # This is to avoid TypedModuleMeta to attempt to
    #   type check _validate, leading to an infinite recursion
    type.__setattr__(TypedModule, "_validate", _validate)



import typing

if typing.TYPE_CHECKING:
    from typing import (
        Annotated as Vmapped,  # noqa: F401
        Annotated as VmappedI,  # noqa: F401
        Annotated as VmappedT,  # noqa: F401
    )
else:
    from ._vmapped import (
        VmappedI as VmappedI,
        VmappedT as VmappedT,
    )

    # Vmapped is the convenient alias for VmappedT
    Vmapped = VmappedT



class TypinoxError(Exception):
    """Base class for all exceptions raised by Typinox."""

    pass


class TypinoxAnnotationError(TypinoxError, TypeError):
    """Raised at annotation-time to indicate an error in the annotation."""

    pass


class TypinoxInvalidTypeToCheck(TypinoxError, TypeError):
    """Raised when the type being checked should not be checked at run-time."""

    pass


class TypinoxTypeViolation(TypinoxError, TypeError):
    """Raised when the typechecking failed."""

    pass


class TypinoxNotImplementedError(TypinoxError, NotImplementedError):
    """Raised when a feature is not yet implemented."""

    pass



import jax.numpy as jnp
import pytest
from beartype import beartype
from jaxtyping import Array, Bool, Float, Scalar, jaxtyped

from typinox import (
    TypedModule,
    set_debug_mode,
)

set_debug_mode(True)


class Foo(TypedModule):
    x: Float[Array, "n m"]

    def equal(self, y: Float[Array, "n m"]) -> Bool[Scalar, ""]:
        return jnp.array_equal(self.x, y)

    def equal_val(self, y: Float[Array, "n m"]) -> Bool[Scalar, ""]:
        self._validate()
        return jnp.array_equal(self.x, y)


@jaxtyped(typechecker=beartype)
def foo_equal(foo: Foo, y: Float[Array, "n m"]) -> Bool[Scalar, ""]:
    return jnp.array_equal(foo.x, y)


def test_module_constraint_scope():
    foo = Foo(x=jnp.zeros((3, 4)))
    assert foo_equal(foo, jnp.zeros((3, 4)))
    assert foo.equal(jnp.zeros((3, 4)))
    assert foo.equal_val(jnp.zeros((3, 4)))

    assert not foo_equal(foo, jnp.ones((3, 4)))
    assert not foo.equal(jnp.ones((3, 4)))
    assert not foo.equal_val(jnp.ones((3, 4)))

    assert not foo_equal(foo, jnp.zeros((3, 5)))
    assert not foo.equal(jnp.zeros((3, 5)))
    with pytest.raises(TypeError, match=r"Float\[Array, 'n m'\]"):
        foo.equal_val(jnp.zeros((3, 5)))



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import sys
from pathlib import Path

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Typinox"
copyright = "2025, Etaoin Wu"
author = "Etaoin Wu"
release = "0.4.0"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.intersphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    # 'autoapi.extension',
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

sys.path.insert(0, str(Path(__file__).resolve().parents[2]))


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_book_theme"
html_static_path = ["_static"]


# -- Intersphinx configuration -----------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#intersphinx

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "beartype": ("https://beartype.readthedocs.io/en/latest/", None),
    "jaxtyping": ("https://docs.kidger.site/jaxtyping/", None),
    "equinox": ("https://docs.kidger.site/equinox/", None),
    "jax": ("https://docs.jax.dev/en/latest/", None),
}



import jax
import pytest
from beartype.door import is_bearable
from jax import (
    numpy as jnp,
    tree as jt,
)
from jaxtyping import Array, Float, Scalar

import typinox as tpx
from typinox import (
    TypedModule,
    ValidatedT,
    ensure_shape,
    ensure_shape_equal,
    set_debug_mode,
)

set_debug_mode(True)


class MatMul(TypedModule):
    a: Float[Array, "m n"]
    b: Float[Array, "n p"]
    c: Float[Array, "m p"]

    def __validate__(self):
        return self.a.shape[0] % 2 == 0

    def residue(self) -> Float[Array, "m p"]:
        return self.a @ self.b - self.c

    def total_residue(self) -> Float[Scalar, ""]:
        return jnp.sum(self.residue() ** 2)


MatMulT = ValidatedT[MatMul]


def test_init_consistent():
    a = jnp.arange(6).reshape(2, 3) / 3 + 1.8
    b = jnp.arange(15).reshape(3, 5) / 7 - 2.4
    c = a @ b
    m = MatMul(a=a, b=b, c=c)
    assert m.total_residue() <= 0.01

    with pytest.raises(TypeError, match="does not match type hint"):
        m = MatMul(a=a, b=b, c=c.T)

    with pytest.raises(TypeError, match="failed its custom validation"):
        m = MatMul(a=b.T, b=a.T, c=c.T)


class MatMulChain(TypedModule):
    a: int = tpx.field(static=True)
    d: int = tpx.field(static=True)
    left: MatMulT
    right: MatMulT
    final: Float[Array, "a d"]

    def __validate__(self):
        ensure_shape_equal(".left.c", self.left.c, ".right.a", self.right.a)
        ensure_shape(".left.a", self.left.a, "a _")
        ensure_shape(".right.b", self.right.b, "_ d")
        ensure_shape(".a", self.a, "a")
        ensure_shape(".d", self.d, "d")

    def total_residue(self) -> Float[Scalar, ""]:
        return (
            self.left.total_residue()
            + self.right.total_residue()
            + jnp.sum((self.left.c - self.right.a) ** 2)
            + jnp.sum((self.final - self.right.c) ** 2)
        )


def meaningless_array(shape):
    size = int(jnp.prod(jnp.array(shape)))
    return jnp.arange(size).reshape(shape) / size - 0.2


def test_fields():
    a = meaningless_array((2, 5))
    b = meaningless_array((5, 4))
    c = a @ b - 1e-5 * meaningless_array((2, 4))
    d = c + 2.3e-6 * meaningless_array((4, 2)).T
    e = meaningless_array((4, 3))
    f = d @ e + 1.7e-5 * meaningless_array((2, 3))
    good_obj = MatMulChain(
        a=2,
        d=3,
        left=MatMul(a=a, b=b, c=c),
        right=MatMul(a=d, b=e, c=f),
        final=f - 1.2e-6 * meaningless_array((3, 2)).T,
    )
    assert good_obj.total_residue() <= 0.01

    good_matmul = good_obj.left
    bad_matmul = jt.map(lambda x: x.T, good_matmul)
    with pytest.raises(TypeError, match="does not match"):
        _ = MatMulChain(
            a=2,
            d=3,
            left=good_matmul,
            right=bad_matmul,
            final=good_obj.final,
        )

    with pytest.raises(TypeError, match="does not match"):
        _ = MatMulChain(
            a=2,
            d=3,
            left=good_matmul,
            right=good_matmul,
            final=good_obj.final,
        )

    bad_f = meaningless_array((2, 4))
    with pytest.raises(TypeError, match="does not match the named dimensions"):
        _ = MatMulChain(
            a=2,
            d=3,
            left=good_obj.left,
            right=good_obj.right,
            final=bad_f,
        )

    with pytest.raises(TypeError, match="does not match the named dimensions"):
        _ = MatMulChain(
            a=2,
            d=4,
            left=good_obj.left,
            right=good_obj.right,
            final=f,
        )


def test_doc_example():
    from typing import Self

    from typinox import ValidationFailed
    from typinox.error import TypinoxTypeViolation

    class SquareMat(TypedModule):
        mat: Float[Array, "n n"]
        n: int = tpx.field(static=True)

        def __validate__(self):
            if self.n != self.mat.shape[0]:
                raise ValidationFailed(
                    f"n={self.n} does not match mat.shape[0]={self.mat.shape[0]}"
                )

        def diagonal_plus_one(self: Self):
            return jnp.diagonal(self.mat) + 1

    f = SquareMat(mat=jnp.eye(3), n=3)
    f.diagonal_plus_one()  # works

    with pytest.raises(TypinoxTypeViolation):
        _ = SquareMat(mat=jnp.eye(3), n=4)  # fails

    f = jax.tree.map(lambda x: x.reshape((9, 1)), f)
    # here f.mat has shape (9, 1) and f.n is 3
    with pytest.raises(TypeError):
        f.diagonal_plus_one()  # fails

    assert isinstance(f, SquareMat)
    assert not is_bearable(f, ValidatedT[SquareMat])  # type: ignore



import chex
import equinox
import jax
import pytest
from jax import (
    numpy as jnp,
    tree as jt,
)

import typinox


@pytest.fixture(autouse=True)
def add_doctest_np(doctest_namespace):
    doctest_namespace["jax"] = jax
    doctest_namespace["jnp"] = jnp
    doctest_namespace["jt"] = jt

    doctest_namespace["typinox"] = typinox
    doctest_namespace["tpx"] = typinox

    doctest_namespace["equinox"] = equinox
    doctest_namespace["eqx"] = equinox

    doctest_namespace["chex"] = chex
